﻿package com.robotacid.dungeon {
	import com.robotacid.geom.Pixel;
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.display.Sprite;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	
	/**
	 * ...
	 * @author steed
	 */
	public class Map {
		
		public var g:Main;
		public var width:int;
		public var height:int;
		public var level:int;
		
		// pacing variables - these keep a consistent scale for the dungeon
		public var vert_pace:int;
		public var horiz_pace:int;
		
		public var rooms:Array/*Room*/;
		public var room:Room;
		
		private var i:int;
		private var j:int;
		private var k:int;
		private var m:int;
		
		public static const MIN_ROOM_WIDTH:int = 3;
		public static const MIN_ROOM_HEIGHT:int = 2;
		
		public static const UP:int = 1;
		public static const RIGHT:int = 2;
		public static const DOWN:int = 4;
		public static const LEFT:int = 8;
		
		public var state:int = 0;
		public var good_level:Boolean = true;
		
		public var grid_width:int;
		public var grid_height:int;
		
		public var cell_width:int;
		public var cell_height:int;
		
		public var exit:Pixel;
		public var entrance:Pixel;
		
		public var neighbours:Vector.<Pixel> = new Vector.<Pixel>(4, true);
		
		public var sub_state:int = 0;
		
		public var map_bitmap:Bitmap;
		
		public function Map(level:int, g:Main) {
			this.g = g;
			this.level = level;
			rooms = [];
			// create pacing standard for this level
			horiz_pace = level * 3;
			vert_pace = level * 2;
			i = 0;
		}
		public function main():void{
			if(state == 0){
				
				// sometimes the generator will fail - then we recreate the dungeon
				good_level = true;
				
				// create a list of rooms, then assign - randomly - 1 siblings
				
				for(i = 0; i < (5 * level) + random(level * 5); i++){
					rooms.push(new Room());
				}
				
				for(i = 0; i < rooms.length; i++){
					do{
						var pick:int = random(rooms.length);
					} while(pick == i);
					rooms[i].siblings.push(rooms[pick]);
				}
				
				// now the hard part, positioning and then connecting them:
				
				// size each room and get a ball park size for each cell in the grid
				
				cell_height = 0;
				cell_width = 0;
				
				
				for(i = 0; i < rooms.length; i++){
					rooms[i].width = MIN_ROOM_WIDTH + random(horiz_pace);
					if(rooms[i].width > cell_width) cell_width = rooms[i].width;
					rooms[i].height = MIN_ROOM_HEIGHT + random(vert_pace);
					if(rooms[i].height > cell_height) cell_height = rooms[i].height;
				}
				
				cell_height += 2;
				cell_width += 2;
				
				// get a grid size for the cells
				// we basically create a minimum of 2 or a maximum of half the number of cells either way
				grid_height = 2 + random((rooms.length / 2) - 2);
				grid_width = Math.ceil((1.0 * rooms.length) / grid_height);
				// now make the grid a little bigger so we get a more eclectic layout
				grid_width++;
				grid_height++;
				
				// let's assign grid numbers
				var nums:Array = [];
				for(i = 0; i < grid_height * grid_width; i++) nums.push(i);
				for(i = 0; i < rooms.length; i++){
					var n:int = random(nums.length);
					rooms[i].grid_num = nums[n];
					nums.splice(n, 1);
				}
				
				// we use a bitmap to work on so we can flood fill and track previous positions:
				map_bitmap = new Bitmap(new BitmapData(grid_width * cell_width, grid_height * cell_height, true, 0xFFFF0000));
				g.addChild(map_bitmap);
				map_bitmap.scaleX *= 2;
				map_bitmap.scaleY *= 2;
				map_bitmap.x = 160 - map_bitmap.width * 0.5;
				map_bitmap.y = 120 - map_bitmap.height * 0.5;
				state++;
				i = 0;
			} else if(state == 1){
				
				// place the rooms in their cells
				rooms[i].x = (rooms[i].grid_num % grid_width) * cell_width;
				rooms[i].y = (rooms[i].grid_num / grid_width);
				rooms[i].y *= cell_height;
				// random offset the rooms
				rooms[i].x += 1 + random((cell_width - 1) - rooms[i].width);
				rooms[i].y += 1 + random((cell_height - 1) - rooms[i].height);
				// paint those rooms
				map_bitmap.bitmapData.fillRect(new Rectangle(
					rooms[i].x, rooms[i].y, rooms[i].width, rooms[i].height
				), 0xFF00FF00);
				i++;
				if(i >= rooms.length){
					state++;
					i = 0;
					j = 0;
					k = 0;
					m = 0;
				}
			} else if(state == 2){
				// now we connect the rooms
				if(m == 0){
					var side:int;
					// pick an exit
					exit = new Pixel();
					room = rooms[i];
					do{
						side = 1 << random(4);
						if(side & UP){
							exit.x = room.x + random(room.width);
							exit.y = room.y;
						} else if(side & RIGHT){
							exit.x = room.x + room.width - 1;
							exit.y = room.y + random(room.height);
						} else if(side & DOWN){
							exit.x = room.x + random(room.width);
							exit.y = room.y + room.height - 1;
						} else if(side & LEFT){
							exit.x = room.x;
							exit.y = room.y + random(room.height);
						}
					} while(room.touchesDoors(exit) || onEdge(exit, map_bitmap.bitmapData.width, map_bitmap.bitmapData.height));
					room.doors.push(exit);
					// now pick an entrance
					entrance = new Pixel();
					room = rooms[i].siblings[j];
					do{
						side = 1 << random(4);
						if(side & UP){
							entrance.x = room.x + random(room.width);
							entrance.y = room.y - 1;
						} else if(side & RIGHT){
							entrance.x = room.x + room.width;
							entrance.y = room.y + random(room.height);
						} else if(side & DOWN){
							entrance.x = room.x + random(room.width);
							entrance.y = room.y + room.height;
						} else if(side & LEFT){
							entrance.x = room.x - 1;
							entrance.y = room.y + random(room.height);
						}
					} while(room.touchesDoors(entrance) || onEdge(entrance, map_bitmap.bitmapData.width, map_bitmap.bitmapData.height));
					room.doors.push(entrance);
				}
				if(m < 200 && (exit.x != entrance.x || exit.y != entrance.y)){
					map_bitmap.bitmapData.setPixel(exit.x, exit.y, 0x00CC00);
					neighbours[0] = new Pixel(exit.x, exit.y - 1);
					neighbours[1] = new Pixel(exit.x + 1, exit.y);
					neighbours[2] = new Pixel(exit.x, exit.y + 1);
					neighbours[3] = new Pixel(exit.x - 1, exit.y);
					var best:int = int.MAX_VALUE;
					var choice:Pixel = null;
					for(var k:int = 0; k < neighbours.length; k++){
						var dist:int = entrance.mDist(neighbours[k]);
						if(map_bitmap.bitmapData.getPixel(neighbours[k].x, neighbours[k].y) != 0x00FF00){
							dist+=3;
						}
						if(dist < best){
							best = dist;
							choice = neighbours[k];
						}
					}
					if(choice){
						exit.x = choice.x;
						exit.y = choice.y;
					} else {
						j++;
						m = 0;
						map_bitmap.bitmapData.setPixel(exit.x, exit.y, 0x00CC00);
						map_bitmap.bitmapData.floodFill(exit.x, exit.y, 0xFF00FF00);
					}
					m++;
				} else {
					j++;
					m = 0;
					map_bitmap.bitmapData.setPixel(exit.x, exit.y, 0x00CC00);
					map_bitmap.bitmapData.floodFill(exit.x, exit.y, 0xFF00FF00);
				}
				if(j >= rooms[i].siblings.length){
					i++;
					j = 0;
				}
				if(i >= rooms.length){
					state++;
				}
			} else if(state == 3){
				
				// did the room generation create two separate networks?
				map_bitmap.bitmapData.floodFill(rooms[0].x, rooms[0].y, 0xFFFFFF00);
				
				for(i = 0; i < rooms.length; i++){
					if(map_bitmap.bitmapData.getPixel32(rooms[i].x, rooms[i].y) != 0xFFFFFF00){
						good_level = false;
					}
				}
				if(!good_level) g.onClick();
				else state++;
				
			} else if(state == 4){
				// trim the map, we may have a large portion of unused rock
				var map_bounds:Rectangle = map_bitmap.bitmapData.getColorBoundsRect(0xFFFFFFFF, 0xFFFFFF00);
				var trimmed_map:Bitmap = new Bitmap(new BitmapData(map_bounds.width + 2, map_bounds.height + 2, true, 0xFFFF0000));
				trimmed_map.bitmapData.copyPixels(map_bitmap.bitmapData, map_bounds, new Point(1, 1));
				map_bitmap.bitmapData = trimmed_map.bitmapData;
			}
		}
		
		/* integer random method - seeing as I'm going to be caning this sort of stuff */
		public static function random(n:int):int{
			return Math.random() * n;
		}
		/* is this pixel sitting on the edge of the map? it will likely cause me trouble if it is... */
		public static function onEdge(pixel:Pixel, width:int, height:int):Boolean{
			return pixel.x<= 0 || pixel.x >= width-1 || pixel.y <= 0 || pixel.y >= height-1;
		}
	}
	
}